@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 # Minio Python Library for Amazon S3 Compatible Cloud Storage, (C)
-# 2015, 2016 Minio, Inc.
+# 2015, 2016, 2017 Minio, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -19,12 +19,15 @@ minio.helpers
 
 This module implements all helper functions.
 
-:copyright: (c) 2015 by Minio, Inc.
+:copyright: (c) 2015, 2016, 2017 by Minio, Inc.
 :license: Apache 2.0, see LICENSE for more details.
 
 """
 
 from __future__ import absolute_import
+# if math.ceil returns an integer and devide two integers returns a float, calculate
+# part size will cause errors, so make sure division integers returns a float.
+from __future__ import division
 import io
 
 import collections
@@ -35,20 +38,47 @@ import os
 import errno
 import math
 
-from .compat import urlsplit, urlencode, str, bytes, basestring
+from .compat import (urlsplit, urlencode, queryencode,
+                     str, bytes, basestring, _is_py3, _is_py2)
 from .error import (InvalidBucketError, InvalidEndpointError,
                     InvalidArgumentError)
 
 # Constants
 MAX_MULTIPART_COUNT = 10000 # 10000 parts
 MAX_MULTIPART_OBJECT_SIZE = 5 * 1024 * 1024 * 1024 * 1024  # 5TiB
-MIN_OBJECT_SIZE = 5 * 1024 * 1024  # 5MiB
+MAX_POOL_SIZE = 10
+MIN_PART_SIZE = 5 * 1024 * 1024  # 5MiB
 
 _VALID_BUCKETNAME_REGEX = re.compile('^[a-z0-9][a-z0-9\\.\\-]+[a-z0-9]$')
 _ALLOWED_HOSTNAME_REGEX = re.compile(
     '^((?!-)[A-Z\\d-]{1,63}(?<!-)\\.)*((?!-)[A-Z\\d-]{1,63}(?<!-))$',
     re.IGNORECASE)
 
+_EXTRACT_REGION_REGEX = re.compile('s3[.-]?(.+?).amazonaws.com')
+
+def get_s3_region_from_endpoint(endpoint):
+    """
+    Extracts and returns an AWS S3 region from an endpoint
+    of form `s3-ap-southeast-1.amazonaws.com`
+
+    :param endpoint: Endpoint region to be extracted.
+    """
+
+    # Extract region by regex search.
+    m = _EXTRACT_REGION_REGEX.search(endpoint)
+    if m:
+        # Regex matches, we have found a region.
+        region = m.group(1)
+        if region == 'external-1':
+            # Handle special scenario for us-east-1 URL.
+            return 'us-east-1'
+        if region.startswith('dualstack'):
+            # Handle special scenario for dualstack URL.
+            return region.split('.')[1]
+        return region
+
+    # No regex matches return None.
+    return None
 
 def dump_http(method, url, request_headers, response, output_stream):
     """
@@ -111,45 +141,45 @@ def mkdir_p(path):
         else:
             raise
 
-
 class PartMetadata(object):
     """
     Parts manager split parts metadata :class:`PartMetadata <PartMetadata>`.
 
     :param data: Part writer object backed by temporary file.
-    :param md5hasher: Computed md5sum Hasher interface.
-    :param sha256hasher: Computed sha256sum Hasher interface.
+    :param md5_hex: MD5 hash in hex format.
+    :param sha256_hex: Sha256 hash in hex format.
     :param size: Size of the part.
     """
-    def __init__(self, data, md5hasher, sha256hasher, size):
+    def __init__(self, data, md5_hex, sha256_hex, size):
         self.data = data
-        self.md5hasher = md5hasher
-        self.sha256hasher = sha256hasher
+        self.md5_hex = md5_hex
+        self.sha256_hex = sha256_hex
         self.size = size
 
-
-def parts_manager(data, part_size=5*1024*1024):
+def read_full(data, size):
     """
-    Reads data and provides temporary files of a given size.
+    read_full reads exactly `size` bytes from reader. returns
+    `size` bytes.
 
-    :param data: Input reader object which needs to be saved.
-    :param part_size: Individual part number defaults to 5MB.
-    :return: Returns :class:`PartMetadata <PartMetadata>`
+    :param data: Input stream to read from.
+    :param size: Number of bytes to read from `data`.
+    :return: Returns :bytes:`part_data`
     """
-    tmpdata = io.BytesIO()
-    md5hasher = Hasher.md5()
-    sha256hasher = Hasher.sha256()
-    total_read = 0
-    while total_read < part_size:
-        current_data = data.read(1024)
+    default_read_size = 32768 # 32KiB per read operation.
+    chunk = io.BytesIO()
+    chunk_size = 0
+
+    while chunk_size < size:
+        read_size = default_read_size
+        if (size - chunk_size) < default_read_size:
+            read_size = size - chunk_size
+        current_data = data.read(read_size)
         if not current_data or len(current_data) == 0:
             break
-        tmpdata.write(current_data)
-        md5hasher.update(current_data)
-        sha256hasher.update(current_data)
-        total_read += len(current_data)
+        chunk.write(current_data)
+        chunk_size+= len(current_data)
 
-    return PartMetadata(tmpdata, md5hasher, sha256hasher, total_read)
+    return chunk.getvalue()
 
 AWS_S3_ENDPOINT_MAP = {
     'us-east-1': 's3.amazonaws.com',
@@ -163,6 +193,7 @@ AWS_S3_ENDPOINT_MAP = {
     'eu-central-1': 's3-eu-central-1.amazonaws.com',
     'ap-south-1': 's3.ap-south-1.amazonaws.com',
     'ap-southeast-1': 's3-ap-southeast-1.amazonaws.com',
+    'ap-southeast-2': 's3-ap-southeast-2.amazonaws.com',
     'ap-northeast-1': 's3-ap-northeast-1.amazonaws.com',
     'ap-northeast-2': 's3-ap-northeast-2.amazonaws.com',
     'cn-north-1': 's3.cn-north-1.amazonaws.com.cn'
@@ -190,10 +221,17 @@ def get_target_url(endpoint_url, bucket_name=None, object_name=None,
     parsed_url = urlsplit(endpoint_url)
 
     # Get new host, scheme.
+    scheme = parsed_url.scheme
     host = parsed_url.netloc
+
+    # Strip 80/443 ports since curl & browsers do not
+    # send them in Host header.
+    if (scheme == 'http' and parsed_url.port == 80) or\
+       (scheme == 'https' and parsed_url.port == 443):
+        host = parsed_url.hostname
+
     if 's3.amazonaws.com' in host:
         host = get_s3_endpoint(bucket_region)
-    scheme = parsed_url.scheme
 
     url = scheme + '://' + host
     if bucket_name:
@@ -217,18 +255,14 @@ def get_target_url(endpoint_url, bucket_name=None, object_name=None,
         ordered_query = collections.OrderedDict(sorted(query.items()))
         query_components = []
         for component_key in ordered_query:
-            if ordered_query[component_key] is not None:
-                if isinstance(ordered_query[component_key], list):
-                    for value in ordered_query[component_key]:
-                        encoded_query = urlencode(
-                            str(value)).replace('/', '%2F')
-                        query_components.append(component_key+'='+encoded_query)
-                else:
-                    encoded_query = urlencode(
-                        str(ordered_query[component_key])).replace('/', '%2F')
-                    query_components.append(component_key+'='+encoded_query)
+            if isinstance(ordered_query[component_key], list):
+                for value in ordered_query[component_key]:
+                    query_components.append(component_key+'='+
+                                            queryencode(value))
             else:
-                query_components.append(component_key)
+                query_components.append(
+                    component_key+'='+
+                    queryencode(ordered_query.get(component_key, '')))
 
         query_string = '&'.join(query_components)
         if query_string:
@@ -264,10 +298,6 @@ def is_valid_endpoint(endpoint):
         if not _ALLOWED_HOSTNAME_REGEX.match(hostname):
             raise InvalidEndpointError('Hostname does not meet URL standards.')
 
-        if hostname.endswith('.amazonaws.com') and \
-           (hostname != 's3.amazonaws.com'):
-            raise InvalidEndpointError('Amazon S3 hostname should be '
-                                       's3.amazonaws.com.')
     except AttributeError as error:
         raise TypeError(error)
 
@@ -289,8 +319,9 @@ def is_virtual_host(endpoint_url, bucket_name):
     # such buckets.
     if 'https' in parsed_url.scheme and '.' in bucket_name:
         return False
-    if 's3.amazonaws.com' in parsed_url.netloc:
-        return True
+    for host in ['s3.amazonaws.com', 'aliyuncs.com']:
+        if host in parsed_url.netloc:
+            return True
     return False
 
 def is_valid_bucket_name(bucket_name):
@@ -338,6 +369,25 @@ def is_non_empty_string(input_string):
 
     return True
 
+def is_valid_policy_type(policy):
+    """
+    Validate if policy is type str
+
+    :param policy: S3 style Bucket policy.
+    :return: True if policy parameter is of a valid type, 'string'.
+    Raise :exc:`TypeError` otherwise.
+    """
+    if _is_py3:
+        string_type = str,
+    elif _is_py2:
+        string_type = basestring
+
+    if not isinstance(policy, string_type):
+        raise TypeError('policy can only be of type str')
+
+    is_non_empty_string(policy)
+
+    return True
 
 def is_valid_bucket_notification_config(notifications):
     """
@@ -469,6 +519,32 @@ def is_valid_bucket_notification_config(notifications):
 
     return True
 
+def is_valid_sse_c_object(sse=None):
+    """
+    Validate the SSE object and type
+
+    :param sse: SSE object defined.
+    """
+    if sse and sse.type() != "SSE-C":
+            raise InvalidArgumentError("Required type SSE-C object to be passed")
+
+def is_valid_sse_object(sse):
+    """
+    Validate the SSE object and type
+
+    :param sse: SSE object defined.
+    """
+    if sse and sse.type() != "SSE-C" and sse.type() != "SSE-KMS" and sse.type() != "SSE-S3":
+        raise InvalidArgumentError("unsuported type of sse argument in put_object")
+
+def is_valid_source_sse_object(sse):
+    """
+    Validate the SSE object and type
+
+    :param sse: SSE object defined.
+    """
+    if sse and sse.type() != "copy_SSE-C":
+        raise InvalidArgumentError("Required type copy_SSE-C object to be passed")
 
 def encode_object_name(object_name):
     """
@@ -481,9 +557,10 @@ def encode_object_name(object_name):
     return urlencode(object_name)
 
 class Hasher(object):
-    """Adaptation of hashlib-based hash functions that return
-    unicode-encoded hex- and base64-digest strings.
-
+    """
+    Adaptation of hashlib-based hash functions that
+    return unicode-encoded hex- and base64-digest
+    strings.
     """
     def __init__(self, data, h):
         if data is None:
@@ -556,8 +633,8 @@ def optimal_part_info(length):
     # Use floats for part size for all calculations to avoid
     # overflows during float64 to int64 conversions.
     part_size_float = math.ceil(length/MAX_MULTIPART_COUNT)
-    part_size_float = (math.ceil(part_size_float/MIN_OBJECT_SIZE)
-                       * MIN_OBJECT_SIZE)
+    part_size_float = (math.ceil(part_size_float/MIN_PART_SIZE)
+                       * MIN_PART_SIZE)
     # Total parts count.
     total_parts_count = int(math.ceil(length/part_size_float))
     # Part size.
@@ -565,3 +642,36 @@ def optimal_part_info(length):
     # Last part size.
     last_part_size = length - int(total_parts_count-1)*part_size
     return total_parts_count, part_size, last_part_size
+
+# return a new metadata dictionary where user defined metadata keys
+# are prefixed by "x-amz-meta-"
+def amzprefix_user_metadata(metadata):
+    m = dict()
+    for k,v in metadata.items():
+       if is_amz_header(k) or is_supported_header(k) or is_storageclass_header(k):
+            m[k] = v
+       else:
+            m["X-Amz-Meta-" + k] = v
+    return m
+
+# returns true if amz s3 system defined metadata
+def is_amz_header(key):
+    key = key.lower()
+    return key.startswith("x-amz-meta") or key == "x-amz-acl" or key.startswith("x-amz-server-side-encryption")
+
+# returns true if a standard supported header
+def is_supported_header(key):
+    ## Supported headers for object.
+    supported_headers = [
+	   "cache-control",
+	   "content-encoding",
+	   "content-disposition",
+	   "content-language",
+	   "x-amz-website-redirect-location",
+            ## Add more supported headers here.
+        ]
+    return key.lower() in supported_headers
+
+# returns true if header is a storage class header
+def is_storageclass_header(key):
+    return key.lower() == "x-amz-storage-class"
